#!/usr/bin/perl

# Attempts to translate grbl-dialect g-code to Marlin g-code as accepted
# by the Snapmaker2, the key word being "attempts".  Check the resulting
# file before you use it to make sure it isn't going to turn your Snapmaker
# into a newt or something else equally bizarre.

use strict;

my $grblfile = $ARGV[0];                    #input file
my $marlinfile = $grblfile . ".marlin.cnc"; #output file
my $metricmult = 1;                         #multiplier for US -> metric

# input grbl and write to marlin, one line at a time
open(GRBL, "<", $grblfile) or die("Couldn't open $grblfile for reading: $!\n");
open(MARLIN, ">", $marlinfile) or die("Couldn't open $marlinfile for writing: $!\n");
while (my $line = <GRBL>){

    # G0 and G1 (positioning/movement)
    if ($line =~ /^(G[01])?([XYZF].+)$/){
        my $command = $1;
        my $incoords = $2;

        my $outcoords = "";
        if ($incoords =~ /X(\-?\d+(\.\d+)?)/){
            my $temp = $1;
            $temp = $temp * $metricmult;
            $outcoords = "$outcoords X$temp";
        }
        if ($incoords =~ /(Y(\-?\d+(\.\d+)?))/){
            my $temp = $2;
            $temp = $temp * $metricmult;
            $outcoords = "$outcoords Y$temp";
        }
        if ($incoords =~ /(Z(\-?\d+(\.\d+)?))/){
            my $temp = $2;
            $temp = $temp * $metricmult;
            $outcoords = "$outcoords Z$temp";
        }
        if ($incoords =~ /(F(\d+(\.\d+)?))/){
            $outcoords = "$outcoords F$2";
        }

        if ($command !~ /G/){
            $command = "G1";
        }

        print MARLIN ($command . $outcoords . "\n");
    }
    elsif ($line =~ /^G0$/){
        #ignore--should be a no-op
    }

    # G4 (dwell)
    elsif ($line =~ /^(G4)/){
        #The Fusion 360 post-process file is somewhat confusing about what syntax is supported.
        my $dwellmsecs = 0;
        if ($line =~ /S(\d+)/){
            $dwellmsecs = $1 * 1000;
        }
        if ($line =~ /P(\d+)/){
            $dwellmsecs = $dwellmsecs + $1;
        }
        if ($dwellmsecs > 99999.999){
            $dwellmsecs = 99999.999;
        }
        if ($dwellmsecs < 0.001){
            $dwellmsecs = 0;
        }

        if ($dwellmsecs > 0) {
            print MARLIN "G4 P$dwellmsecs\n";
        }
    }

    # G17, G18, G19 (workspace planes)
    elsif ($line =~ /^(G1[789])/){
        print("Ignoring unsupported workspace plane command.\n");
    }

    # G20, G21 (select units)
    elsif ($line =~ /^(G21)/){
        #metric (millimeter) units
        $metricmult = 1;
        print MARLIN "G21\n";
    }
    elsif ($line =~ /^(G20)/){
        #loon attempting to use inch units
        $metricmult = 25.4;
    }

    # G28, G28.2 (homing)
    elsif ($line =~ /^(G28)/){
        print("Replacing homing command with G0 X0 Y0 Z0 (hopefully that's what was intended).\n");
        print MARLIN ";HOMING\n";
        print MARLIN "G0 X0 Y0 Z0\n";
    }

    # G2 and G3 (arcs)
    elsif ($line =~ /^(G[23])/){
        my $direction = $1;

        #There are two possible formats for this monstrosity.  One designates the center of the
        #arc using an offset from the starting point, the other uses a radius.  Currently we
        #can only handle the offset type.
        if ($line =~ /R\d/){
            print("Unsupported arc command in radius format!\n");
        }
        $line =~ /X(\d+(\.\d+)?)/;
        my $coords = "X$1";
        if ($line =~ /Y(\d+(\.\d+)?)/){
            $coords = "$coords Y$1";
        }
        if ($line =~ /Z(\d+(\.\d+)?)/){
            $coords = "$coords Z$1";
        }
        $line =~ /I(-?\d+(\.\d+)?)/;
        my $center = "I$1";
        $line =~ /J(-?\d+(\.\d+)?)/;
        $center = "$center J$1";

        print MARLIN "$direction $coords $center\n";
    }

    # G90, G91, G91.1 (distance mode)
    elsif ($line =~ /^(G9[01])/){
        print("Passing distance mode command straight through.\n");
        print MARLIN "$line\n";
    }

    # M0, M1, M2, M30 (pause/end)
    elsif (($line =~ /^(M[012])$/) or ($line =~ /^(M30)/)){
        print("Ignoring unsupported pause or end command (probably harmless).\n");
    }

    # M6 (toolchange)
    elsif ($line =~ /^(M0*6)/){
        print("This file contains a toolchange!  The Snapmaker2 does not support toolchanges at this time.\n");
        print("If you want to split the file manually at the toolchange, search for the line that starts with ;TOOLCHANGE\n");
        print MARLIN ";TOOLCHANGE $line";
    }

    # M3, M5 (start/stop spindle)
    elsif ($line =~ /^(M0*3)/){
        #Snapmaker spindles operate in percentages.  This is somewhat unusual for CNC.
        #Convert on the basis that 100% = 12000RPM
        $line =~ /M0*3 ?S(\d+)/;
        my $rpm = $1;
        my $percent = ($rpm / 12000) * 100;
        if ($percent > 100) {
            $percent = 100;
        }

        print MARLIN "M3 S$percent\n";
    }
    elsif ($line =~ /^(M0*5)/){
        print MARLIN "M5\n";
    }

    else {
        print("Unknown command $line");
    }

}
close GRBL;
close MARLIN;
print ("Done!  Please inspect $marlinfile and the above output for anything that looks weird or incomplete.\n");
